# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementations of quantum gates.

The :class:`.GateImplementation` subclasses in this subpackage construct :class:`.TimeBox` instances to
implement specific native gates, using the calibration data that the class has been initialized with.
Each GateImplementation instance encapsulates the calibration data for a specific implementation of a specific
native gate acting on a specific locus.

Several different implementations and calibration schemes can be supported for a given gate,
each represented by its own GateImplementation subclass.
Likewise, a single GateImplementation subclass can be sometimes used to implement several different gates
through different calibration data.
"""

import copy
from dataclasses import replace

from iqm.pulse.gate_implementation import GateImplementation
from iqm.pulse.gates.barrier import Barrier
from iqm.pulse.gates.conditional import (
    CCPRX_Composite,
    CCPRX_Composite_DRAGCosineRiseFall,
    CCPRX_Composite_DRAGGaussian,
)
from iqm.pulse.gates.cz import (
    CZ_CRF,
    CZ_CRF_ACStarkCRF,
    CZ_GaussianSmoothedSquare,
    CZ_Slepian,
    CZ_Slepian_ACStarkCRF,
    CZ_Slepian_CRF,
    CZ_TruncatedGaussianSmoothedSquare,
    FluxPulse_SmoothConstant_SmoothConstant,
    FluxPulseGate_CRF_CRF,
    FluxPulseGate_TGSS_CRF,
)
from iqm.pulse.gates.default_gates import _implementation_library, _quantum_ops_library
from iqm.pulse.gates.delay import Delay
from iqm.pulse.gates.flux_multiplexer import FluxMultiplexer_SampleLinear
from iqm.pulse.gates.measure import Measure_Constant, Measure_Constant_Qnd, Shelved_Measure_Constant
from iqm.pulse.gates.move import MOVE_CRF_CRF, MOVE_SLEPIAN_CRF, MOVE_TGSS_CRF
from iqm.pulse.gates.prx import (
    Constant_PRX_with_smooth_rise_fall,
    PRX_DRAGCosineRiseFall,
    PRX_DRAGCosineRiseFallSX,
    PRX_DRAGGaussian,
    PRX_DRAGGaussianSX,
    PRX_FastDrag,
    PRX_FastDragSX,
    PRX_HdDrag,
    PRX_HdDragSX,
    PRX_ModulatedDRAGCosineRiseFall,
    get_unitary_prx,
)
from iqm.pulse.gates.reset import Reset_Conditional, Reset_Wait
from iqm.pulse.gates.rz import (
    RZ_ACStarkShift_CosineRiseFall,
    RZ_ACStarkShift_smoothConstant,
    RZ_PRX_Composite,
    RZ_Virtual,
    get_unitary_rz,
)
from iqm.pulse.gates.sx import SXGate
from iqm.pulse.gates.u import UGate, get_unitary_u
from iqm.pulse.quantum_ops import QuantumOp, QuantumOpTable

_exposed_implementations: dict[str, type[GateImplementation]] = {
    cls.__name__: cls
    for cls in (
        Barrier,
        Constant_PRX_with_smooth_rise_fall,
        PRX_DRAGGaussian,
        PRX_DRAGCosineRiseFall,
        PRX_DRAGGaussianSX,
        PRX_DRAGCosineRiseFallSX,
        PRX_FastDragSX,
        PRX_FastDrag,
        PRX_HdDrag,
        PRX_HdDragSX,
        SXGate,
        UGate,
        RZ_PRX_Composite,
        RZ_Virtual,
        CZ_CRF_ACStarkCRF,
        CZ_Slepian_ACStarkCRF,
        CZ_GaussianSmoothedSquare,
        CZ_Slepian,
        CZ_Slepian_CRF,
        CZ_CRF,
        FluxPulse_SmoothConstant_SmoothConstant,
        CZ_TruncatedGaussianSmoothedSquare,
        FluxPulseGate_TGSS_CRF,
        FluxPulseGate_CRF_CRF,
        Measure_Constant,
        Measure_Constant_Qnd,
        Shelved_Measure_Constant,
        PRX_ModulatedDRAGCosineRiseFall,
        MOVE_CRF_CRF,
        MOVE_SLEPIAN_CRF,
        MOVE_TGSS_CRF,
        RZ_ACStarkShift_CosineRiseFall,
        RZ_ACStarkShift_smoothConstant,
        CCPRX_Composite,
        CCPRX_Composite_DRAGCosineRiseFall,
        CCPRX_Composite_DRAGGaussian,
        Reset_Conditional,
    )
}
"""These GateImplementations can be referred to in the configuration YAML."""


def get_implementation_class(class_name: str) -> type[GateImplementation] | None:
    """Get gate implementation class by class name."""
    return _exposed_implementations.get(class_name)


def expose_implementation(implementation: type[GateImplementation], overwrite: bool = False) -> None:
    """Add the given gate implementation to the list of known implementations.

    Args:
        implementation: GateImplementation to add so that it can be found with :func:`.get_implementation_class`.
        overwrite: If True, does not raise an error if implementation already exists.

    """
    name = implementation.__name__
    if name in _exposed_implementations:
        if not overwrite and _exposed_implementations[name] is not implementation:
            raise ValueError(f"GateImplementation '{name}' has already been defined.")
    _exposed_implementations[name] = implementation


def _validate_implementation(
    op_name: str,
    impl_name: str,
    impl_class_name: str,
) -> None:
    """Check that canonical implementation names cannot be overridden.

    Args:
        op_name: Name of the operation.
        impl_name: Name of the new implementation.
        impl_class_name: Name of the GateImplementation class it maps to.

    Raises:
        ValueError: A canonical implementation name is being redefined.

    """
    default_implementations = _implementation_library.get(op_name, {})

    # check if the implementation name is canonical for this op
    if (impl_class := default_implementations.get(impl_name)) is not None:
        if impl_class_name != impl_class.__name__:
            raise ValueError(
                f"'{op_name}': '{impl_name}' is a reserved implementation name that refers to "
                f"'{impl_class.__name__}', and cannot be overridden. "
                "Consider renaming your implementation."
            )


def register_operation(
    operations: dict[str, QuantumOp],
    op: QuantumOp,
    overwrite: bool = False,
) -> None:
    """Register a new QuantumOp to the given operations table.

    Args:
        operations: Known operations, to which the new operation is added.
        op: Definition for the new operation.
        overwrite: If True, allows replacing an existing operation in ``operations``.

    Raises:
        ValueError: ``op.name`` exists in ``operations`` and ``overwrite==False``.
        ValueError: ``op.name`` is the name of a canonical operation in iqm-pulse.

    """
    if op.name in operations and not overwrite:
        raise ValueError(f"'{op.name}' already registered.")
    if op.name in _quantum_ops_library:
        raise ValueError(f"'{op.name}' conflicts with a canonical operation in iqm-pulse. Use a different name.")

    # make a deep copy since the dicts inside QuantumOp are mutable
    operations[op.name] = copy.deepcopy(op)


def register_implementation(
    operations: dict[str, QuantumOp],
    op_name: str,
    impl_name: str,
    impl_class: type[GateImplementation],
    *,
    set_as_default: bool = False,
    overwrite: bool = False,
) -> None:
    """Register a new GateImplementation.

    Args:
        operations: Known operations, to which the new implementation is added.
        op_name: Name of the operation under which the implementation is registered.
        impl_name: Name of the implementation to register.
        impl_class: Implementation class to register.
        set_as_default: Whether to set as the default implementation for ``op_name``.
        overwrite: If True, allows replacing an existing implementation.

    Raises:
        ValueError: ``op_name`` does not exist in ``operations``.
        ValueError: The implementation exists and ``overwrite==False``.


    """
    if (op := operations.get(op_name)) is None:
        raise ValueError(f"Operation '{op_name}' is not known, register it first.")

    # canonical implementation names must not be overridden
    _validate_implementation(op_name, impl_name, impl_class.__name__)

    # only overwrite existing implementations with permission
    if (old_class := op.implementations.get(impl_name)) is not None and not overwrite:
        if old_class is not impl_class:
            # cannot change an existing implementation name
            raise ValueError(f"'{op_name}' already has an implementation named '{impl_name}'.")
    else:
        # add the new implementation
        op.implementations[impl_name] = impl_class

    if set_as_default:
        op.set_default_implementation(impl_name)

    if get_implementation_class(impl_class.__name__) is None:
        expose_implementation(impl_class, overwrite=overwrite)



# Copyright 2025 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Force a delay between instructions on the control channels of specific locus components.

Ideally the delay corresponds to an identity gate. In reality it of course allows decoherence
to act on the quantum state for some time.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from iqm.pulse.gate_implementation import GateImplementation
from iqm.pulse.playlist.channel import round_duration_to_granularity_samples
from iqm.pulse.playlist.instructions import Wait
from iqm.pulse.playlist.schedule import Schedule

if TYPE_CHECKING:
    from iqm.pulse.timebox import TimeBox


DELAY_MAX_DURATION_SECONDS = 0.1
"""Maximum duration for individual Delay operations, in seconds."""


class Delay(GateImplementation):
    r"""Applies a delay on the control channels of its locus components.

    This operation applies :class:`.Wait` instructions on all the control channels of all its locus
    components.  The duration of all the Waits is the same, and it is given as a parameter for the
    operation, rounded up to the nearest possible duration the hardware can handle.

    .. note::

       We can only guarantee that the delay is *at least* of the requested duration.  Also, when
       Delay is used in a quantum circuit, the delay between the preceding and following operations is
       again *at least* the requested duration, but could be much more depending on the other operations
       in the circuit.  To see why, consider e.g. the circuit
       ``[CZ(a, b), Delay(1, a), Delay(10, b), CZ(a, b)]`` where a and b are qubits.
       In this case the actual delay between the two CZ gates will be 10 time units rounded up to
       hardware granularity.
    """

    symmetric: bool = True

    def _call(self, duration: float) -> TimeBox:  # type: ignore[override]
        """Delay instruction.

        Args:
            duration: Duration of the requested wait (in seconds). Will be rounded up to the nearest
                duration that the hardware enables, with the exception that a duration of zero will
                cause no waiting. However, as usual, during scheduling all channels the TimeBox is
                acting on will be extended to the duration of the longest channel in the TimeBox.

        """
        if duration > DELAY_MAX_DURATION_SECONDS:
            raise ValueError(
                f"Requested delay duration {duration} s exceeds the allowed maximum {DELAY_MAX_DURATION_SECONDS} s."
                " You can use several delay operations in a row if this is not enough."
            )
        channel_names = self.builder.get_control_channels(self.locus)
        channels = [self.builder.channels[name] for name in channel_names]

        duration_samples: int = (
            0
            if duration == 0.0
            else round_duration_to_granularity_samples(channels, duration, round_up=True, force_min_duration=True)
        )
        # NOTE we assume that all the control channels here have the same sample rate!
        wait = Wait(duration_samples)
        timebox = self.to_timebox(Schedule({ch: [wait] for ch in channel_names}))
        timebox.neighborhood_components[0] = set(self.locus)
        return timebox



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Canonical quantum operations and implementations provided by iqm-pulse."""

from __future__ import annotations

from typing import TYPE_CHECKING, Final

import numpy as np

from iqm.pulse.gates.barrier import Barrier
from iqm.pulse.gates.conditional import (
    CCPRX_Composite,
    CCPRX_Composite_DRAGCosineRiseFall,
    CCPRX_Composite_DRAGGaussian,
)
from iqm.pulse.gates.cz import (
    CZ_CRF,
    CZ_CRF_ACStarkCRF,
    CZ_GaussianSmoothedSquare,
    CZ_Slepian,
    CZ_Slepian_ACStarkCRF,
    CZ_Slepian_CRF,
    CZ_TruncatedGaussianSmoothedSquare,
    FluxPulseGate_CRF_CRF,
    FluxPulseGate_TGSS_CRF,
)
from iqm.pulse.gates.delay import Delay
from iqm.pulse.gates.flux_multiplexer import FluxMultiplexer_SampleLinear
from iqm.pulse.gates.measure import Measure_Constant, Shelved_Measure_Constant
from iqm.pulse.gates.move import MOVE_CRF_CRF, MOVE_SLEPIAN_CRF, MOVE_TGSS_CRF
from iqm.pulse.gates.prx import (
    PRX_DRAGCosineRiseFall,
    PRX_DRAGCosineRiseFallSX,
    PRX_DRAGGaussian,
    PRX_DRAGGaussianSX,
    PRX_ModulatedDRAGCosineRiseFall,
    get_unitary_prx,
)
from iqm.pulse.gates.reset import Reset_Conditional, Reset_Wait
from iqm.pulse.gates.rz import (
    RZ_ACStarkShift_CosineRiseFall,
    RZ_PRX_Composite,
    RZ_Virtual,
    get_unitary_rz,
)
from iqm.pulse.gates.sx import SXGate
from iqm.pulse.gates.u import UGate, get_unitary_u
from iqm.pulse.quantum_ops import QuantumOp

if TYPE_CHECKING:
    from iqm.pulse.gate_implementation import GateImplementation


_implementation_library: dict[str, dict[str, type[GateImplementation]]] = {
    "barrier": {"": Barrier},
    "delay": {"wait": Delay},
    "measure": {
        "constant": Measure_Constant,
    },
    "measure_fidelity": {
        "constant": Measure_Constant,
        "shelved_constant": Shelved_Measure_Constant,
    },
    "prx": {
        "drag_gaussian": PRX_DRAGGaussian,
        "drag_crf": PRX_DRAGCosineRiseFall,
        "drag_crf_sx": PRX_DRAGCosineRiseFallSX,
        "drag_gaussian_sx": PRX_DRAGGaussianSX,
    },
    "prx_12": {
        "modulated_drag_crf": PRX_ModulatedDRAGCosineRiseFall,
    },
    "u": {"prx_u": UGate},
    "sx": {"prx_sx": SXGate},
    "rz": {"virtual": RZ_Virtual, "prx_composite": RZ_PRX_Composite},
    "rz_physical": {"ac_stark_crf": RZ_ACStarkShift_CosineRiseFall},
    "cz": {
        "tgss": CZ_TruncatedGaussianSmoothedSquare,
        "tgss_crf": FluxPulseGate_TGSS_CRF,
        "crf_crf": FluxPulseGate_CRF_CRF,
        "crf": CZ_CRF,
        "gaussian_smoothed_square": CZ_GaussianSmoothedSquare,
        "slepian": CZ_Slepian,
        "slepian_crf": CZ_Slepian_CRF,
        "crf_acstarkcrf": CZ_CRF_ACStarkCRF,
        "slepian_acstarkcrf": CZ_Slepian_ACStarkCRF,
    },
    "move": {"tgss_crf": MOVE_TGSS_CRF, "crf_crf": MOVE_CRF_CRF, "slepian_crf": MOVE_SLEPIAN_CRF},
    "cc_prx": {
        "prx_composite": CCPRX_Composite,
        "prx_composite_drag_crf": CCPRX_Composite_DRAGCosineRiseFall,
        "prx_composite_drag_gaussian": CCPRX_Composite_DRAGGaussian,
    },
    "reset": {"reset_conditional": Reset_Conditional},
    "reset_wait": {"reset_wait": Reset_Wait},
    "flux_multiplexer": {"sample_linear": FluxMultiplexer_SampleLinear},
}
"""For each canonical quantum operation name, maps its canonical implementation implementation names
to their GateImplementation classes.

Canonical names are reserved, and the users cannot redefine them.
"""

_quantum_ops_library = {
    op.name: op
    for op in [
        QuantumOp(
            "barrier",
            0,
            implementations=_implementation_library["barrier"],
            symmetric=True,
        ),
        QuantumOp(
            "delay",
            0,
            ("duration",),
            implementations=_implementation_library["delay"],
            symmetric=True,
        ),
        QuantumOp(
            "measure",
            0,
            ("key",),
            implementations=_implementation_library["measure"],
            factorizable=True,
        ),
        QuantumOp(
            "measure_fidelity",
            0,
            ("key",),
            implementations=_implementation_library["measure_fidelity"],
            factorizable=True,
        ),
        QuantumOp(
            "prx",
            1,
            ("angle", "phase"),
            implementations=_implementation_library["prx"],
            unitary=get_unitary_prx,
        ),
        QuantumOp(
            "prx_12",
            1,
            ("angle", "phase"),
            implementations=_implementation_library["prx_12"],
        ),
        QuantumOp(
            "u",
            1,
            ("theta", "phi", "lam"),
            implementations=_implementation_library["u"],
            unitary=get_unitary_u,
        ),
        QuantumOp(
            "sx",
            1,
            implementations=_implementation_library["sx"],
            unitary=lambda: get_unitary_prx(np.pi / 2, 0),
        ),
        QuantumOp(
            "rz",
            1,
            ("angle",),
            implementations=_implementation_library["rz"],
            unitary=get_unitary_rz,
        ),
        QuantumOp(
            "rz_physical",
            1,
            implementations=_implementation_library["rz_physical"],
        ),
        QuantumOp(
            "cz",
            2,
            (),
            implementations=_implementation_library["cz"],
            symmetric=True,
            unitary=lambda: np.diag([1.0, 1.0, 1.0, -1.0]),
        ),
        QuantumOp(
            "move",
            2,
            implementations=_implementation_library["move"],
        ),
        QuantumOp(
            "cc_prx",
            1,
            ("angle", "phase", "feedback_qubit", "feedback_key"),
            implementations=_implementation_library["cc_prx"],
        ),
        QuantumOp(
            "reset",
            0,
            implementations=_implementation_library["reset"],
            symmetric=True,
            factorizable=True,
        ),
        QuantumOp(
            "reset_wait",
            0,
            implementations=_implementation_library["reset_wait"],
            symmetric=True,
            factorizable=True,
        ),
        QuantumOp(
            "flux_multiplexer",
            0,
            implementations=_implementation_library["flux_multiplexer"],
        ),
    ]
}
"""Canonical quantum operations provided by iqm-pulse.

Their names are reserved, and the users cannot redefine them.
"""

_deprecated_ops: Final[set[str]] = set()
"""Names of canonical quantum operations that are deprecated.

They are not included by default in ScheduleBuilder unless the user specifically requests them."""
_deprecated_implementations: Final[dict[str, set[str]]] = {}
"""For each canonical quantum operation name, canonical implementation names that are deprecated.

They are not included by default in ScheduleBuilder unless the user specifically requests them."""
# TODO: deprecate gaussian_smoothed_square and everything with the tgss waveform as that is considered inferior to crf.
# TODO: deprecate PRX_drag_gaussian
